/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#define NDEBUG

#include "../config.h"

#include "LocalizationMUSIC.hpp"

#include <libharkio/libharkio.h> 

#include <iostream>
#include <fstream>

#include <boost/bind.hpp>
#include "evd.hpp"
#include "svd.hpp"

using namespace std;
using namespace boost;
using namespace boost::numeric;
using namespace harkio;

struct less_for_vec {
  bool operator()(vector<float>& v, int i, int j) { return v[i] < v[j]; }
};

LocalizationMUSIC::LocalizationMUSIC(int num_mic, int sampling_rate,
                                     int tlength, int shift) :
  NumMicrophones(num_mic), TimeLength(tlength), Shift(shift),
  SamplingRate(sampling_rate), R(num_mic, num_mic), V(num_mic, num_mic),
  RN(num_mic, num_mic), VN(num_mic, num_mic),
  TempCalc1(num_mic, num_mic), TempCalc2(num_mic, num_mic),
  TempCalc3(num_mic, num_mic), TempCalc4(num_mic, num_mic),
  TempCalc5(num_mic, num_mic), TempCalc6(num_mic, num_mic),
  TempCalc7(num_mic, num_mic), TempCalc8(num_mic, num_mic),
  TempCalc9(num_mic, num_mic),
  NumSource(num_mic)
{
  FreqLength = TimeLength / 2 + 1;

  UsesFreq.resize(FreqLength);

  EigenValue.resize(FreqLength);
  CEigenValue.resize(NumMicrophones);
  REigenValue.resize(NumMicrophones);
  MaxRxx.resize(FreqLength);
  MaxRxxN.resize(FreqLength);  
  E.resize(FreqLength);
  EN.resize(FreqLength);
  for (int k = 0; k < FreqLength; k++) {
    EigenValue[k].resize(NumMicrophones);
    E[k].resize(NumMicrophones, NumMicrophones);
    EN[k].resize(NumMicrophones, NumMicrophones);
  }

  FoundDirection.resize(NumSource);
  FoundRange.resize(NumSource);

}

void LocalizationMUSIC::SetUsedFrequency(int lo, int up)
{
  int lower = (int)((float)lo * TimeLength / SamplingRate);
  int upper = (int)((float)up * TimeLength / SamplingRate);
  for (int k = lower; k < upper; k++) {
    UsesFreq[k] = true;
  }
}

void LocalizationMUSIC::Localize(vector<fcmatrix>& Rxx, vector<fcmatrix>& RxxN, 
				int mindeg, int maxdeg, 
				string music_alg, 
				ublas::vector<float>& MSSW,
				bool enable_eigenvalue_weight)
{
  for (int k = 0; k < UsesFreq.size(); k++) {

    MaxRxx[k] = 1.0;
    MaxRxxN[k] = 1.0;    

    if (!UsesFreq[k]) {
      continue;
    }

    float max = MaxOfAbsValue(Rxx[k]);
    noalias(R) = Rxx[k] / (max + 1.0f);
    MaxRxx[k] = max + 1.0f;
    
    if(music_alg != "SEVD"){
      float maxN = MaxOfAbsValue(RxxN[k]);
      noalias(RN) = RxxN[k] / (maxN);
      MaxRxxN[k] = maxN;
    }

    if(music_alg == "SEVD"){
      evd(R, EigenValue[k], V);
    }
    else if (music_alg == "GEVD"){
      invMAT(RN,TempCalc1);
      evd(TempCalc1, REigenValue, VN);
      for(int i = 0; i < REigenValue.size(); i++) REigenValue[i] = pow(REigenValue[i], (float)0.5);    
      invMAT(VN,TempCalc2);
      for(int i = 0; i < TempCalc3.size1(); i++) {
	for(int j = 0; j < TempCalc3.size2(); j++) {
	  TempCalc3(i,j) = 0.0;
	  if(i==j) TempCalc3(i,j) = REigenValue[i];
	}
      }
      noalias(TempCalc6) = prod(TempCalc3,TempCalc2);
      noalias(TempCalc4) = prod(VN, TempCalc6);
      noalias(TempCalc7) = prod(R,TempCalc4);
      noalias(TempCalc5) = prod(TempCalc4, TempCalc7);
      evd(TempCalc5, EigenValue[k], V);
    }
    else if (music_alg == "GSVD"){
      invMAT(RN,TempCalc1);
      noalias(TempCalc2) = prod(TempCalc1, R);
      svd(TempCalc2, REigenValue, TempCalc4, TempCalc3);
      sort_eig_fvec_ascending(REigenValue, TempCalc4, EigenValue[k], V);
    }

    noalias(E[k]) = V;    
        
  }

  //printf("\n");
    
  CalcAveragePower(EigenValue, E, A, UsesFreq, NumSource, Pavg, Pavg_bin, MSSW, enable_eigenvalue_weight);

  for (int id = 0; id < SearchDirection.size(); id++) {
    Pavg_sum[id] = 0.0f;
    for (int ir = 0; ir < SearchRange.size(); ir++) {
      Pavg_sum[id] = Pavg[id][ir];
    }
    Pavg_sum[id] = 10 * log10(Pavg_sum[id] + 1.0f);
  }

  FoundSourceCount = SearchLocation(Pavg_sum, Pavg, NumSource,
                                    FoundDirection, FoundRange, SearchDirection, mindeg, maxdeg);
  
}

void LocalizationMUSIC::ReadTransferFunction(const char* filename, 
					       vector<int>& ch)
{

  if (HARK::checkIdentifier(filename) == HARKIO_TRUE) {
    HARK hark_read;
    hark_read.readFile(filename);

    cerr << "Try to read " << filename << " as HARK format." << endl;

    Chunk::VERSION_CHECK version_check;
    FMT* fmt = hark_read.getData()->getChunk<FMT>();
    // TODO: version check
    if (fmt->getFormatId() != FMT::M2PG) {
      // TODO: 顼
    }

    M2PG* m2pg = hark_read.getData()->getChunk<M2PG>();
    int slength = fmt->getNFFT() / 2 + 1;
    int nmic = fmt->getNMic();
    int nmic2 = ch.size();
    for (int c = 0; c < ch.size(); c++) {
      if (!(ch[c] >= 0 && ch[c] < nmic)) {
	cerr << "LocalizationMUSIC. Nonexistent channel is selected." << endl;
	exit(1);
      }
    }

    SearchRange.resize(m2pg->getNR());
    A.resize(extents[m2pg->getND()][m2pg->getNR()]);
    for (int id = 0; id < m2pg->getND(); id++) {
      for (int ir = 0; ir < m2pg->getNR(); ir++) {
        A[id][ir].resize(slength, nmic2);
      }
    }
    M2PG::SRC_LIST_TYPE* src = m2pg->getSrc();

    SearchDirection.resize(m2pg->getND());
    for (int id = 0; id < m2pg->getND(); id++) {
      SearchDirection[id] = (int)round(atan2((*src)[id][0].y,
                                             (*src)[id][0].x) * 180 / M_PI);
    }

    for (int ir = 0; ir < m2pg->getNR(); ir++) {
      SearchRange[ir] = (int)sqrt((*src)[0][ir].x * (*src)[0][ir].x +
                                  (*src)[0][ir].y * (*src)[0][ir].y +
                                  (*src)[0][ir].z * (*src)[0][ir].z);
    }

    M2PG::TF_LIST_TYPE* tf = m2pg->getTF();
    for (int id = 0; id < m2pg->getND(); id++) {
      for (int ir = 0; ir < m2pg->getNR(); ir++) {
        for (int m = 0; m < nmic2; m++) {
          for (int k = 0; k < slength; k++) {
            A[id][ir](k, m) = (*tf)[id][ir][ch[m]][k];
          }
        }
      }
    }
  }
  else {
    cerr << harkio::Error::getInstance().getMsg() << endl;
    cerr << "Try to read " << filename << " as header-less MUSIC transfer function format.\n";
    ::ReadTransferFunction(filename, A, SearchDirection, SearchRange, ch);
  }

  Pavg_sum.resize(SearchDirection.size());
  Pavg.resize(SearchDirection.size());
  for (int id = 0; id < SearchDirection.size(); id++) {
    Pavg[id].resize(SearchRange.size());
  }
  Pavg_bin.resize(A[0][0].size1());
  for (int bins = 0; bins < A[0][0].size1(); bins++) {
    Pavg_bin[bins].resize(SearchDirection.size());
  }
  
  cerr << SearchDirection.size() << " directions, "
       << SearchRange.size() << " ranges, "
       << A[0][0].size2() << " microphones, "
       << (A[0][0].size1() - 1) * 2 << " points" << endl;

}

void LocalizationMUSIC::CalcAveragePower(vector<fvector>& w, vector<fcmatrix>& E,
					AMatrixArray& A,
					vector<bool>& UsesFreq,
					int num_source, 
					vector<vector<float> >& Pavg,
					vector<vector<float> >& Pavg_bin,
					ublas::vector<float>& Weight,
					bool enable_eigenvalue_weight)
{
  int nd = Pavg.size();
  int nr = Pavg[0].size();
  int num_mic = w[0].size();
    
  for (int id = 0; id < nd; id++) {
    for (int ir = 0; ir < nr; ir++) {
      Pavg[id][ir] = 0.0f;
    }
  }

  for (int row = 0; row < Pavg_bin.size(); row++) {
    for (int col = 0; col < Pavg_bin[0].size(); col++) {
      Pavg_bin[row][col] = 0.0f;
    }
  }

  float G = 0.0f;
  float norm_a = 0.0f;
  for (int k = 0; k < UsesFreq.size(); k++) {
    if (!UsesFreq[k]) {
      continue;
    }
        
    for (int id = 0; id < nd; id++) {
      for (int ir = 0; ir < nr; ir++) {
        G = 0.0f;
        for (int m = num_mic - num_source - 1; m >= 0; m--) {
          G += norm(ublas::inner_prod(ublas::herm(ublas::row(A[id][ir], k)),
                                      ublas::column(E[k], m)));
        }
        norm_a = ublas::inner_prod(ublas::herm(ublas::row(A[id][ir], k)),
                                   ublas::row(A[id][ir], k)).real();
	float EV_Weight = (enable_eigenvalue_weight ? sqrt(w[k][num_mic - 1]) : 1.0f );
        Pavg[id][ir] += (norm_a / G) * Weight[k] * EV_Weight;
	cerr << "";
     }
      Pavg_bin[k][id] = 10 * log10((norm_a / G * Weight[k]) + 1.0f);
    }
  }
}

int LocalizationMUSIC::SearchLocation(vector<float>& Pavg_sum, vector<vector<float> >& Pavg,
				     int n_source, vector<int>& direction, vector<int>& range, 
				     vector<int>& SDirection, int mindeg, int maxdeg)
{
  vector<float> fSDirection(SDirection.size());
  for(int i = 0; i < SDirection.size(); i++){
    fSDirection[i] = (float)SDirection[i];
  }
  return SearchLocation(Pavg_sum, Pavg, n_source,
			direction, range, fSDirection, mindeg, maxdeg);
}

int LocalizationMUSIC::SearchLocation(vector<float>& Pavg_sum, vector<vector<float> >& Pavg,
				     int n_source, vector<int>& direction, vector<int>& range, 
				     vector<float>& SDirection, int mindeg, int maxdeg)
{
  int n_result;
  int nd = Pavg.size();
  int nr = Pavg[0].size();
  vector<int> candidate;
  vector<float> tmp(nd + nr);
  float tmp_min;
  float tmp_max;

  candidate.reserve(nd + nr);

  candidate.resize(0);
  SearchPeak(Pavg_sum, candidate, SDirection, mindeg, maxdeg);

  if (candidate.size() < n_source) {
    n_result = candidate.size();
  }
  else {
    n_result = n_source;
  }

  partial_sort(candidate.begin(),
               candidate.begin() + n_result,
               candidate.end(),
               bind<bool>(less_for_vec(), Pavg_sum, _2, _1));
  copy(candidate.begin(), candidate.begin() + n_result, direction.begin());

  candidate.resize(0);
  for (int source = 0; source < n_result; source++) {
    int id = direction[source];
    SearchPeak(Pavg[id], candidate, SDirection, -180, 180);
    vector<int>::iterator it =
      max_element(candidate.begin(), candidate.end(),
                  bind<bool>(less_for_vec(), Pavg_sum, _1, _2));
    range[source] = *it;
  }

  return n_result;
}

void LocalizationMUSIC::SearchPeak(vector<float>& power, vector<int>& peak_cand, 
				   vector<float>& SDirection, int mindeg, int maxdeg)
{
  float p1;
  float p2;
  float p3;
  int deg;

  while( maxdeg - mindeg > 360 ) maxdeg -= 360;
  while( maxdeg - mindeg <= 0 ) maxdeg += 360;
    
  if (power.size() < 2) {
    peak_cand.push_back(0);
  }
  else {
    for (int i = 0; i < power.size(); i++) {
      deg = (int)(SDirection[i] + 0.5);
      while( deg - mindeg > 360 ) deg -= 360;
      while( deg - mindeg <= 0 )  deg += 360;
                  
      if ( deg > mindeg && deg < maxdeg ){
        p1 = (i == 0 ? power.back() : power[i - 1]);
        p2 = power[i];
        p3 = (i == power.size() - 1 ? power.front() : power[i + 1]);
	
        if (p2 - p1 > 0 && p3 - p2 < 0) {
	  peak_cand.push_back(i);
        }
      }
    }
  }
}

